/*
  Copyright (c) 2011-12 Andreas Paffenholz
 
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version: http://www.gnu.org/licenses/gpl.txt.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
-------------------------------------------------------------------------------
  $Project: toric_varieties $$Id: polytopal.cc 210 2012-04-11 15:47:27Z paffenholz $
*/

#include "polymake/client.h"
#include "polymake/Matrix.h"
#include "polymake/Array.h"
#include "polymake/SparseMatrix.h"
#include "polymake/Integer.h"
#include "polymake/Rational.h"
#include "polymake/common/lattice_tools.h"
#include "polymake/common/ntl_wrappers.h"

namespace polymake { namespace fan {

    // decide whether a fan is the face fan (equivalently: normal fan) of a polytope
    // we use the method of Theorem 4.8 in Chapter V of Ewald
    // Let K be the Gale dual of the rays of the fan
    // and C_j the polytope spanned by the colums of the Gale dual whose index 
    // corresponds to rays not in the maximal cone sigma_j of the fan
    // The fan is polytopal if and only if there is a point in the intersection of the 
    // relative interiors of the polytopes C_j
    // 
    // we use a linear programming approach:
    // if b+A_jx\ge 0, l+L_jx=0 is an outer description of C_j,
    // then we consider the linear system of equations and inequalities
    // given by b+A_jx-s\ge 0, l+L_jx=0 for all j
    // if this has a feasible solution with s>0 then the fan is polytopal
    //
    // FIXME alternative method: 
    // the (d-1)-skeleton of the fan determines the edge directions of the polytope, if there is one
    // assign a variable x_j to each direction 
    // the fan is polytopal if and only if
    // there is a positive solution for the x_j that satisfies
    // a closing condition around any 2-face (i.e. (d-2)-face of the cone): the weighted edge lengths should sum to 0

    bool polytopal ( perl::Object fan ) {

      Matrix<Rational> rays = fan.give("RAYS");
      Array<Set<int> > max_cones = fan.give("MAXIMAL_CONES");
      Matrix<Integer> t_int_rays=T(common::primitive(rays));
      Matrix<Integer> ker = common::integer_kernel(t_int_rays,false);

      // FIXME ch conversion should be done by calling the appropriate cdd function directly
      perl::Object c("Polytope<Rational>");      
      Matrix<Integer> Pt = T(ker.minor(All,~max_cones[0]));
      Pt = (ones_vector<Integer>(Pt.rows()))|Pt;
      c.take("INPUT_RAYS") << Pt;

      Matrix<Rational> L = c.give("LINEAR_SPAN");
      Matrix<Rational> At = c.give("FACETS");
      Matrix<Rational> A = At|-(ones_vector<Rational>(At.rows()));

      for ( int i = 1; i < max_cones.size(); ++i ) {	

	// FIXME ch conversion should be done by calling the appropriate cdd function directly
	perl::Object c2("Polytope<Rational>");
	Matrix<Integer> Pt = T(ker.minor(All,~max_cones[i]));
	Pt = (ones_vector<Integer>(Pt.rows()))|Pt;
	c2.take("INPUT_RAYS") << Pt;
	Matrix<Rational> Lt = c2.give("LINEAR_SPAN");
	L /= Lt;

	c2.give("FACETS") >> At;
	A /= At|(-(ones_vector<Rational>(At.rows())));
      }

      L = L|(zero_vector<Rational>(L.rows()));

      // FIXME lp solution should be computed by calling the appropriate cdd function directly
      perl::Object p("polytope::Polytope<Rational>");
      p.take("INEQUALITIES") << A;
      p.take("EQUATIONS") << L;
      
      perl::Object lp("LinearProgram");
      lp.take("LINEAR_OBJECTIVE") << unit_vector<Rational>(A.cols(),A.cols()-1);
      p.add("LP",lp);

      Rational max = p.give("LP.MAXIMAL_VALUE");

      if ( max > 0 ) 
	return true;

      return false;
    }

    Function4perl(&polytopal, "polytopal($)");

  }}
